local saveButton = "F2"
local loadButton = "F3"

local PLUGIN 		= {}
local SAVE 			= {}
local IMP_LOGICS 	= require "Plugins.QuickSavesLogics"
local objectsList = {}

local saveSwitch, loadSwitch, gameSaved = nil, nil, nil
local inGameTime = ffi.cast("int*", 0x535928)

local function _chamAdd(addr, code)
	local cham_exe = ffi.cast("char*",addr)
	local i = 0 	
	for v in string.gmatch(code, "([^ ]+)") do
		cham_exe[i] = tonumber(v, 16)
		i = i+1
	end
end

_chamAdd(0x4C976D, "E9 2F EE F6 FF") -- jmp [Chameleon90] ; when objects are getting destroyed
_chamAdd(0x4385A1, "56 52 89 FE BA 5A 00 00 00 E8 BD 39 0D 00 5A 5E A1 00 B0 5A 00 E9 B7 11 09 00")
--[[
Chameleon90 =
	push esi
	push edx
	mov esi, edi ; object to destroy
	mov edx, 9
	call [50bf6c] ; main Chameleon
	pop edx
	pop esi
	mov eax, dword ptr ds:[5AB000]
	jmp [4c9772]
]]

PLUGIN.ClawAnimation = {
	[4] 	= "CLAW_JUMP",
	[8] 	= "CLAW_WALK",
	[9] 	= "CLAW_WALK",
	[25] 	= "CLAW_FALL",
	[27] 	= "CLAW_CLIMB",
	[5002] 	= "CLAW_LAND",
	[5006] 	= "CLAW_TEETER",
	[5007] 	= "CLAW_SWING"
}

PLUGIN.RespawnObject = function(data)
	local object = CreateObject{
		X 			= data.X,
		Y 			= data.Y,
		Z			= data.Z,
		Smarts 		= data.Smarts,
		Powerup 	= data.Powerup,
		SpeedX 		= data.SpeedX,
		SpeedY 		= data.SpeedY,
		XMin 		= data.XMin,
		XMax 		= data.XMax,
		YMin 		= data.YMin,
		YMax 		= data.YMax,
		UserRect1 	= data.UserRect1,
		UserRect2 	= data.UserRect2,
		Logic 		= data.Logic,
		Image 		= data.Image
	}
	return object
end

PLUGIN.SaveObject = function(id, object)
	if object ~= nil and object._v ~= nil then
		local logicName = IMP_LOGICS[ffi.cast("int*", object._v)[4]]
		if logicName then
			--MessageBox(id.." "..GetImgStr(object.Image).." "..object.X.." "..object.Y.." "..object.State)
			SAVE.Objects[id] = {
				X 			= object.X,
				Y 			= object.Y,
				Z			= object.Z,
				Smarts 		= object.Smarts,
				Health 		= object.Health,
				Powerup 	= object.Powerup,
				SpeedX 		= object.SpeedX,
				SpeedY 		= object.SpeedY,
				XMin 		= object.XMin,
				XMax 		= object.XMax,
				YMin 		= object.YMin,
				YMax 		= object.YMax,
				UserRect1 	= ffi.new("Rect", object.UserRect1.Left, object.UserRect1.Top, object.UserRect1.Right, object.UserRect1.Bottom),
				UserRect2 	= ffi.new("Rect", object.UserRect2.Left, object.UserRect2.Top, object.UserRect2.Right, object.UserRect2.Bottom),
				Logic 		= logicName,
				Image 		= GetImgStr(object.Image),
				State 		= object.State
			}
		end
	end
end

PLUGIN.SaveGame = function()
	if GetClaw().Health <= 0 or GetClaw().State == ClawStates.Hit then
		TextOut"Can't save now!"
		return
	end
	SAVE.Time = GetTime()
	SAVE.Claw = {
		X 						= GetClaw().X,
		Y 						= GetClaw().Y,
		Z						= GetClaw().Z,
		I						= GetClaw().I,
		DrawFlags				= GetClaw().DrawFlags.flags,
		Health 					= GetClaw().Health,
		Score 					= GetClaw().Score,
		State					= GetClaw().State
	}
	SAVE.Player = {
		Direction				= PData().Direction,
		RunningSpeedTime 		= PData().RunningSpeedTime,
		PistolAmmo 				= PData().PistolAmmo,
		MagicAmmo 				= PData().MagicAmmo,
		DynamiteAmmo 			= PData().DynamiteAmmo,
		Lives 					= PData().Lives,
		CollectedCoin 			= PData().CollectedCoin,
        CollectedGoldbar 		= PData().CollectedGoldbar,
        CollectedRing 			= PData().CollectedRing,
        CollectedChalice 		= PData().CollectedChalice,
        CollectedCross 			= PData().CollectedCross,
        CollectedScepter 		= PData().CollectedScepter,
        CollectedGecko 			= PData().CollectedGecko,
        CollectedCrown 			= PData().CollectedCrown,
        CollectedSkull 			= PData().CollectedSkull,
        GameCollectedCoin 		= PData().GameCollectedCoin,
        GameCollectedGoldbar 	= PData().GameCollectedGoldbar,
        GameCollectedRing 		= PData().GameCollectedRing,
        GameCollectedChalice 	= PData().GameCollectedChalice,
        GameCollectedCross 		= PData().GameCollectedCross,
        GameCollectedScepter 	= PData().GameCollectedScepter,
        GameCollectedGecko 		= PData().GameCollectedGecko,
        GameCollectedCrown 		= PData().GameCollectedCrown,
        GameCollectedSkull 		= PData().GameCollectedSkull
	}
	SAVE.RespawnPointX		= GetRespawnPoint().X
	SAVE.RespawnPointY		= GetRespawnPoint().Y
	SAVE.Powerup 			= _CurrentPowerup[0]
	SAVE.PowerupTime 		= _PowerupTime[0]
	SAVE.Boss				= GetBoss()
	SAVE.Objects 			= {}
	SAVE.Cannons			= {}
	for id, object in pairs(objectsList) do
		PLUGIN.SaveObject(id, object)
	end
	gameSaved = true
end

PLUGIN.LoadGame = function()
	if not gameSaved then
		TextOut"Quick save does not exist!"
		return
	end
	TextOut" "
	SetCameraPoint(SAVE.Claw.X, SAVE.Claw.Y)
	GetClaw().Health 				= SAVE.Claw.Health
	GetClaw().Score 				= SAVE.Claw.Score
	GetClaw().State					= SAVE.Claw.State
	local ani = PLUGIN.ClawAnimation[SAVE.Claw.State] or "CLAW_STAND"
	GetClaw():SetAnimation(ani)
	GetClaw().DrawFlags.flags		= SAVE.Claw.DrawFlags
	GetClaw().Flags.NoHit			= false
	GetClaw().Flags.Safe			= false
	GetClaw().Z						= SAVE.Claw.Z
	for name, value in pairs(SAVE.Player) do
		PData()[name] = value
	end
	PData().AttemptNb 				= PData().AttemptNb + 1 -- resets CrumblingPeg and BreakPlank
	for id, data in pairs(SAVE.Objects) do
		local currentObject = objectsList[id]
		if currentObject == nil then
			--MessageBox("Case 1:\n"..id.." "..data.X.." "..data.Y.." "..data.Logic.." "..data.State)
			objectsList[id] = PLUGIN.RespawnObject(data)
			if SAVE.Cannons[id] then
				ffi.cast("ObjectA*", SAVE.Cannons[id]):Destroy()
				SAVE.Cannons[id] = nil
			end
		elseif data.Logic:sub(-10) == "Checkpoint" and data.State ~= currentObject.State then
			--MessageBox("Case 2:\n"..id.." "..data.X.." "..data.Y.." "..data.Logic.." "..data.State.."\n"..currentObject.State)
			currentObject:Destroy()
			objectsList[id] = PLUGIN.RespawnObject(data)
		elseif data.Logic == "BreakPlank" then
			-- Just do nothing
		elseif data.Logic == "Officer" or data.Logic == "Soldier" or data.Logic == "Rat" or 
		data.Logic == "CutThroat" or data.Logic == "RobberThief" or data.Logic == "TownGuard1" or 
		data.Logic == "TownGuard2" or data.Logic == "Seagull" or data.Logic == "BearSailor" or 
		data.Logic == "RedTailPirate" or data.Logic == "HermitCrab" or data.Logic == "CrazyHook" or
		data.Logic == "PegLeg" or data.Logic == "Fish" or data.Logic == "Siren" or
		data.Logic == "Chameleon" or data.Logic == "TigerGuard" or data.Logic == "Mercat" and currentObject.State >= 2000 then
			--MessageBox("Case 3:\n"..id.." "..data.X.." "..data.Y.." "..data.Logic.." "..data.State.."\n"..currentObject.State)
			currentObject:Destroy()
			objectsList[id] = PLUGIN.RespawnObject(data)
		elseif (data.State == 5 or data.State == 0) and currentObject.State > 5 then
			--MessageBox("Case 4:\n"..id.." "..data.X.." "..data.Y.." "..data.Logic.." "..data.State.."\n"..currentObject.State)
			currentObject:Destroy()
			objectsList[id] = PLUGIN.RespawnObject(data)
			if SAVE.Cannons[id] then
				ffi.cast("ObjectA*", SAVE.Cannons[id]):Destroy()
				SAVE.Cannons[id] = nil
			end
		elseif currentObject.State >= 2000 then
			--MessageBox("Case 5:\n"..id.." "..data.X.." "..data.Y.." "..data.Logic.." "..data.State.."\n"..currentObject.State)
			currentObject.Health = data.Health
		end
	end
	for id, object in pairs(objectsList) do
		local savedObjData = SAVE.Objects[id]
		if not savedObjData and object ~= nil and object._v ~= nil then
			local logicName = IMP_LOGICS[ffi.cast("int*", object._v)[4]]
			if logicName then
				object:DestroyGlitter()
				object:Destroy()
			end
		end
	end
	snRes(SAVE.RespawnPointX, 11, 17)
	snRes(SAVE.RespawnPointY, 11, 18)
	ClawGivePowerup(SAVE.Powerup, 100)
	_PowerupTime[0] = SAVE.Powerup
	SetBoss(SAVE.Boss)
	inGameTime[0] = SAVE.Time
	Teleport(SAVE.Claw.X, SAVE.Claw.Y)
	CameraToClaw()
end

PLUGIN.GameplayLoop = function(obj)

	-- Remove treasures once collected and enemies once defeated from the objects list:
	if obj.State == 9001 or obj.State == 2004 or obj.Logic == CrabNest and obj.State == 2070 then
		local id = table.key(objectsList, obj)
		if id then
			objectsList[id] = nil
		end
	end
	-- Mirror PunkRat cannon when it respawns:
	if obj.Logic == PunkRat then
		local id = table.key(objectsList, obj)
		if id and tonumber(ffi.cast("int", obj._userdata)) ~= 0 then
			local cannon = obj._userdata[268]
			ffi.cast("ObjectA*", cannon).DrawFlags.flags = obj.DrawFlags.flags
			ffi.cast("ObjectA*", cannon).SpeedX = obj.DrawFlags.Mirror == true and -500 or 500
			if not SAVE.Cannons[id] then
				SAVE.Cannons[id] = cannon
			end
		end
	end
	-- Add dropped items, crabs, crates and amulet gems to the list:
	if obj.Logic == BouncingGoodie or obj.Logic == HermitCrab or obj.Logic == Crate or obj.Logic == EndOfLevelPowerup then
		local id = table.key(objectsList, obj)
		if not id then
			table.insert(objectsList, #objectsList+1, obj)
		end
	end
end

PLUGIN.map = function(ptr)
	-- Init:
	if _chameleon[0] == chamStates.LoadingAssets then
		SAVE = {Claw = {}, Objects = {}, Cannons = {}}
		objectsList, saveSwitch, loadSwitch, gameSaved = {}, nil, nil, nil
	end
	-- Save all objects to a list:
	if _chameleon[0] == chamStates.LoadingObjects then
		local addr = tonumber(ffi.cast("int", ptr))
		if addr and addr > 0 then
			local object = ffi.cast("ObjectA*", addr)
			table.insert(objectsList, object)
		end
	end
	if _chameleon[0] == chamStates.Gameplay then
		LoopThroughObjects(PLUGIN.GameplayLoop)
		-- Save button:
		if KeyPressed(saveButton) and not saveSwitch then
			saveSwitch = true
			PLUGIN.SaveGame()
			if gameSaved then
				TextOut"Quick save activated!"
				PlaySound"GAME_NIPUP2"
			end
		elseif not KeyPressed(saveButton) then
			saveSwitch = false
		end
		-- Load button:
		if KeyPressed(loadButton) and not loadSwitch then
			loadSwitch = true
			PLUGIN.LoadGame()
		elseif KeyPressed(loadButton) then
			loadSwitch = false
		end
	end
	-- When destroying objects:
	if _chameleon[0] == 90 then
		local addr = tonumber(ffi.cast("int", ptr))
		if addr and addr > 0 then
			local object = ffi.cast("ObjectA*", addr)
			local id = table.key(objectsList, object)
			if id then
				objectsList[id] = nil
			end
		end
	end
end

return PLUGIN